
import pkg_resources
import queue
import nose
import threading
import unittest

from pathspider.base import SHUTDOWN_SENTINEL
from pathspider.observer import Observer

class ChainTestCase(unittest.TestCase):

    def setUp(self):
        try:
            import plt # libtrace may not be available
        except ImportError:
            raise nose.SkipTest

    def create_observer(self, test_trace, chains):
        if not test_trace.startswith("/"):
            test_trace = pkg_resources.resource_filename("pathspider",
                                                         "tests/data/" +
                                                         test_trace)
        self.lturi = "pcap:" + test_trace
        self.observer = Observer(self.lturi, chains)
        self.flowqueue = queue.Queue()

    def run_observer(self):
        self.observer_thread = threading.Thread(target=self.observer.run_flow_enqueuer,
                                                args=(self.flowqueue,),
                                                daemon=True)
        self.observer_thread.start()

        flows = []

        while True:
            f = self.flowqueue.get()
            if f == SHUTDOWN_SENTINEL:
                break
            flows.append(f)

        return flows

    def tearDown(self):
        self.observer_thread.join(3)
        assert not self.observer_thread.is_alive()




import argparse
import sys
import logging

import pathspider.cmd.filter
import pathspider.cmd.measure
import pathspider.cmd.observe
import pathspider.cmd.test

cmds = [
    pathspider.cmd.filter,
    pathspider.cmd.measure,
    pathspider.cmd.observe,
    pathspider.cmd.test,
]

def handle_args(argv):
    class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
        def _format_action(self, action):
            parts = super()._format_action(action)
            if action.nargs == argparse.PARSER:
                parts = "\n".join([line for line in parts.split("\n")[1:]])
                parts += "\n\nSpider safely!"
            return parts

    parser = argparse.ArgumentParser(description=('PATHspider will spider the '
                                                  'paths.'),
                                     formatter_class=SubcommandHelpFormatter)

    parser.add_argument("--verbose", action="store_true",
                        help="Enable verbose logging")

    # Add commands
    subparsers = parser.add_subparsers(title="Commands",
                                       metavar='COMMAND', help='command to run')

    # Register commands arguments
    for cmd in cmds:
        cmd.register_args(subparsers)

    args = parser.parse_args(argv[1:])

    # Set up logging
    logging.basicConfig()
    if args.verbose:
        logging.getLogger().setLevel(logging.DEBUG)
    else:
        logging.getLogger().setLevel(logging.INFO)

    # If it's a valid command, run it, or help the user if not
    if hasattr(args, 'cmd'):
        args.cmd(args)
    else:
        parser.print_help()

def handle_args_wrapper():
    handle_args(sys.argv)

if __name__ == "__main__":
    handle_args_wrapper()



"""
.. module:: pathspider.chains.dscp
   :synopsis: A flow analysis chain for Differentiated Services

This module contains the DSCPChain flow analysis chain which can be used by
PATHspider's Observer for recording Differentiated Services [RFC2474]_ details.

.. codeauthor:: Iain R. Learmonth <irl@fsfe.org>

"""

from pathspider.chains.base import Chain
from pathspider.chains.tcp import TCP_SYN

class DSCPChain(Chain):
    """
    This flow analysis chain records details of the Differentiated Services
    Field in the IP header.

    +-----------------------+------+------------------------------------------------+
    | Field Name            | Type | Meaning                                        |
    +=======================+======+================================================+
    | ``dscp_mark_syn_fwd`` | int  | The value of the Differentiated Services       |
    |                       |      | codepoint seen on a TCP SYN packet in the      |
    |                       |      | forward direction                              |
    +-----------------------+------+------------------------------------------------+
    | ``dscp_mark_data_fwd``| int  | The value of the Differentiated Services       |
    |                       |      | codepoint seen on a non-TCP packet or a TCP    |
    |                       |      | packet with a payload in the forward direction |
    +-----------------------+------+------------------------------------------------+
    | ``dscp_mark_syn_rev`` | int  | The value of the Differentiated Services       |
    |                       |      | codepoint seen on a TCP SYN packet in the      |
    |                       |      | reverse direction                              |
    +-----------------------+------+------------------------------------------------+
    | ``dscp_mark_data_rev``| int  | The value of the Differentiated Services       |
    |                       |      | codepoint seen on a non-TCP packet or a TCP    |
    |                       |      | packet with a payload in the reverse direction |
    +-----------------------+------+------------------------------------------------+
    """

    def new_flow(self, rec, ip):
        """
        For a new flow, all fields will be initialised to ``None``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: Always ``True``
        :rtype: bool
        """

        rec['dscp_mark_syn_fwd'] = None
        rec['dscp_mark_syn_rev'] = None
        rec['dscp_mark_data_fwd'] = None
        rec['dscp_mark_data_rev'] = None
        return True

    def ip4(self, rec, ip, rev):
        """
        Records DSCP markings from an IPv4 header.

        DSCP Marking
            For the first TCP SYN packet and the first non-TCP packet or TCP
            packet with a payload, the DSCP value will be recorded in the
            relevant field.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv4 packet that was observed to be part of this flow
        :type ip: plt.ip
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: Always ``True``
        :rtype: bool
        """

        return self._dscp_extract(rec, ip, rev)

    def ip6(self, rec, ip, rev):
        """
        Records DSCP markings from an IPv6 header.

        DSCP Marking
            For the first TCP SYN packet and the first non-TCP packet or TCP
            packet with a payload, the DSCP value will be recorded in the
            relevant field.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv6 packet that was observed to be part of this flow
        :type ip: plt.ip6
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: Always ``True``
        :rtype: bool
        """

        return self._dscp_extract(rec, ip, rev)

    def _dscp_extract(self, rec, ip, rev):
        tos = ip.traffic_class
        dscp = tos >> 2
    
        if ip.tcp:
            if ip.tcp.flags & TCP_SYN == TCP_SYN:
                rec['dscp_mark_syn_rev' if rev else 'dscp_mark_syn_fwd'] = dscp
                return True
            if ip.tcp.payload is None:
                return True
    
        # If not TCP or TCP non-SYN
        data_key = 'dscp_mark_data_rev' if rev else 'dscp_mark_data_fwd'
        rec[data_key] = rec[data_key] or dscp
        return True



"""
.. module:: pathspider.chains.evilbit
   :synopsis: A flow analysis chain for EvilBit

This module contains the EvilChain analysis chain which can be used by
PATHspider's Observer for recording Evil Bit connectivity [RFC3514] details.

"""

from pathspider.chains.tcp import TCP_SYN
from pathspider.chains.base import Chain

class EvilChain(Chain):

    """
    
    +-----------------------+------+------------------------------------------------+
    | Field Name            | Type | Meaning                                        |
    +=======================+======+================================================+
    | ``evilbit_syn_fwd``   | bool | True if the evil bit was set in the IP header  |
    |                       |      | for a TCP SYN packet in the forward direction, |
    |                       |      | false otherwise                                |
    +-----------------------+------+------------------------------------------------+
    | ``evilbit_syn_fwd``   | bool | True if the evil bit was set in the IP header  |
    |                       |      | for a TCP SYN packet in the reverse direction, |
    |                       |      | false otherwise                                |
    +-----------------------+------+------------------------------------------------+
    | ``evilbit_data_fwd``  | bool | True if the evil bit was set in the IP header  |
    |                       |      | for a non-TCP packet in the forward direction, |
    |                       |      | false otherwise                                |
    +-----------------------+------+------------------------------------------------+
    | ``evilbit_data_rev``  | bool | True if the evil bit was set in the IP header  |
    |                       |      | for a non-TCP packet in the  reverse direction,|
    |                       |      | false otherwise                                |
    +-----------------------+------+------------------------------------------------+
    """

    def new_flow(self, rec, ip):
        """
        For a new flow, all fields will be initialised to ``None``.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IP or IPv6 packet that triggered the creation of a new
                   flow record
        :type ip: plt.ip or plt.ip6
        :return: Always ``True``
        :rtype: bool
        """

        rec['evilbit_syn_fwd'] = None
        rec['evilbit_syn_rev'] = None
        rec['evilbit_data_fwd'] = None
        rec['evilbit_data_rev'] = None
        return True
    
    def ip4(self, rec, ip, rev):
        """
        Records evil bit markings from an IPv4 header.

        Evil Bit Marking
            For either TCP_SYN packets or non-TCP or TCP with payload packets
            the relevant field will record whether the Evil Bit was set.

        :param rec: the flow record
        :type rec: dict
        :param ip: the IPv4 packet that was observed to be part of this flow
        :type ip: plt.ip
        :param rev: True if the packet was in the reverse direction, False if
                    in the forward direction
        :type rev: bool
        :return: Always ``True``
        :rtype: bool
        """
        evil = ip.has_rf
            
        if ip.tcp:
            if ip.tcp.flags & TCP_SYN == TCP_SYN:
                rec['evilbit_syn_rev' if rev else 'evilbit_syn_fwd'] = evil
                return True
            if ip.tcp.payload is None:
                return True
            
        # If not TCP or TCP non-SYN
        data_key = 'evilbit_data_rev' if rev else 'evilbit_data_fwd'
        if rec[data_key] is None:
            rec[data_key] = evil
            return True





import pathspider
from pathspider.base import PluggableSpider
from pathspider.single import SingleSpider
from pathspider.helpers.http import connect_http
from pathspider.helpers.http import connect_https
from pathspider.helpers.tcp import connect_tcp
from pathspider.chains.basic import BasicChain
from pathspider.chains.mss import MSSChain
from pathspider.chains.tcp import TCPChain

class MSS(SingleSpider, PluggableSpider):

    name = "mss"
    description = "TCP Maximum Segment Size"
    version = pathspider.base.__version__
    chains = [BasicChain, TCPChain, MSSChain]
    connect_supported = ["tcp", "http", "https", "dnstcp"]

    def combine_flows(self, flows):
        conditions = []

        if not flows[0]['observed']:
            return ['pathspider.not_observed']

        if flows[0]['tcp_connected']:
            conditions.append('mss.connectivity.online')
            conditions.append('mss.option.local.value:' + str(flows[0]['mss_value_fwd']))
            if flows[0]['mss_len_rev'] is not None:
                conditions.append('mss.option.remote.value:' + str(flows[0]['mss_value_rev']))
                if (flows[0]['mss_value_rev'] < flows[0]['mss_value_fwd']):
                    conditions.append('mss.option.received.deflated')
                elif (flows[0]['mss_value_rev'] == flows[0]['mss_value_fwd']):
                    conditions.append('mss.option.received.unchanged')
                else:
                    conditions.append('mss.option.received.inflated')
            else:
                conditions.append('mss.option.received.absent')
        else:
            conditions.append('mss.connectivity.offline')

        return conditions




import nose

import pathspider.base
import pathspider.cmd.measure

def test_plugin_load():
    try:
        # attempt to load dependencies of plugins
        import pldns
    except ImportError:
        raise nose.SkipTest

    expected_names = set(['ECN', 'DSCP', 'UDPZero', 'DNSResolv', 'H2', 'EvilBit', 'MSS'])
    names = set()

    for plugin in pathspider.cmd.measure.plugins:
        assert issubclass(plugin, pathspider.base.Spider)
        names.add(plugin.__name__)

    assert names == expected_names




import argparse
import logging
import json
import sys
import threading
import csv

from straight.plugin import load

from pathspider.base import PluggableSpider
from pathspider.base import SHUTDOWN_SENTINEL
from pathspider.network import interface_up

plugins = load("pathspider.plugins", subclasses=PluggableSpider)

def job_feeder_ndjson(inputfile, spider):
    logger = logging.getLogger("feeder")
    seen_targets = set()
    with open(inputfile) as fh:
        logger.debug("job_feeder: started")
        for line in fh:
            try:
                job = json.loads(line)
                if 'dip' in job.keys():
                    if job['dip'] in seen_targets:
                        logger.debug("This target has already had a job submitted, skipping.")
                        continue
                    seen_targets.add(job['dip'])
                spider.add_job(job)
            except ValueError:
                logger.warning("Unable to decode JSON for a job, skipping...")

        logger.info("job_feeder: all jobs added, waiting for spider to finish")
        spider.shutdown()
        logger.debug("job_feeder: stopped")

def job_feeder_csv(inputfile, spider):
    logger = logging.getLogger("feeder")
    seen_targets = set()
    with open(inputfile) as csvfile:
        reader = csv.reader(csvfile)
        logger.debug("job_feeder: started")
        for row in reader:
            if len(row) == 2:
                job = {'rank': row[0], 'domain': row[1]}
            elif len(row) == 3: # used in 0.9.x release series
                job = {'dip': row[0], 'dp': row[1], 'domain': row[2]}
            elif len(row) == 4: # used in 1.0.x release series
                job = {'dip': row[0], 'dp': row[1], 'domain': row[2], 'rank': row[3]}
            else:
                logger.warning("Unable to read row for a job, skipping...")
                continue

            if 'dip' in job.keys():
                if job['dip'] in seen_targets:
                    logger.debug("This target has already had a job submitted, skipping.")
                    continue
                seen_targets.add(job['dip'])

            spider.add_job(job)

        logger.info("job_feeder: all jobs added, waiting for spider to finish")
        spider.shutdown()
        logger.debug("job_feeder: stopped")




def run_measurement(args):
    logger = logging.getLogger("pathspider")

    try:
        if hasattr(args, "spider"):
            if interface_up(args.interface):
                spider = args.spider(args.workers, "int:" + args.interface, args)
            else:
                logger.error("The chosen interface is not up! Cannot continue.")
                sys.exit(1)
        else:
            logger.error("Plugin not found! Cannot continue.")
            logger.error("Use --help to list all plugins.")
            sys.exit(1)

        logger.info("activating spider...")

        spider.start()

        logger.debug("starting job feeder...")
        if args.csv_input:
            job_feeder = job_feeder_csv
        else:
            job_feeder = job_feeder_ndjson

        threading.Thread(target=job_feeder, args=(args.input, spider)).start()

        with open(args.output, 'w') as outputfile:
            logger.info("opening output file "+args.output)
            while True:
                result = spider.outqueue.get()
                if result == SHUTDOWN_SENTINEL:
                    logger.info("output complete")
                    break
                if not args.output_flows:
                    result.pop("flow_results", None)
                    result.pop("missed_flows", None)
                outputfile.write(json.dumps(result) + "\n")
                logger.debug("wrote a result")
                spider.outqueue.task_done()

    except KeyboardInterrupt:
        logger.error("Received keyboard interrupt, dying now.")

def register_args(subparsers):
    class SubcommandHelpFormatter(argparse.RawDescriptionHelpFormatter):
        def _format_action(self, action):
            parts = super()._format_action(action)
            if action.nargs == argparse.PARSER:
                parts = "\n".join([line for line in parts.split("\n")[1:]])
                parts += "\n\nSpider safely!"
            return parts

    parser = subparsers.add_parser(name='measure',
                                   help="Perform a PATHspider measurement",
                                   formatter_class=SubcommandHelpFormatter)
    parser.add_argument('-i', '--interface', default="eth0",
                        help="The interface to use for the observer. (Default: eth0)")
    parser.add_argument('-w', '--workers', type=int, default=20,
                        help="Number of workers to use. (Default: 20)")
    parser.add_argument('--input', default='/dev/stdin', metavar='INPUTFILE',
                        help=("A file containing a list of PATHspider jobs. "
                              "Defaults to standard input."))
    parser.add_argument('--csv-input', action='store_true',
                        help=("Indicate CSV format."))
    parser.add_argument('--output', default='/dev/stdout', metavar='OUTPUTFILE',
                        help=("The file to output results data to. "
                              "Defaults to standard output."))
    parser.add_argument('--output-flows', action='store_true',
                        help="Include flow results in output.")

    # Set the command entry point
    parser.set_defaults(cmd=run_measurement)

    # Add plugins
    plugin_subparsers = parser.add_subparsers(title="Plugins",
                                              description="The following plugins are available for use:",
                                              metavar='PLUGIN', help='plugin to use')
    for plugin in plugins:
        plugin.register_args(plugin_subparsers)




from pathspider.tests.chains import ChainTestCase

from pathspider.chains.basic import BasicChain

class TestBasicChain(ChainTestCase):

    # TODO: Add tests for UDP/SCTP/UDP-lite/DCCP

    def test_chain_basic_ipv4_tcp(self):
        test_trace = "basic_ipv4_tcp.pcap"
        self.create_observer(test_trace, [BasicChain])

        expected_basic = {
            'oct_fwd': 1747,
            'dp': 80,
            'pkt_rev': 32,
            'pkt_fwd': 32,
            'proto': 6,
            'oct_rev': 41030,
            'sip': '139.133.208.62',
            'sp': 38878,
            'dip': '139.133.1.4'
        }

        flows = self.run_observer()
        print(flows)
        assert len(flows) == 1

        for key in expected_basic:
            assert flows[0][key] == expected_basic[key]

    def test_chain_basic_ipv6_tcp(self):
        test_trace = "basic_ipv6_tcp.pcap"
        self.create_observer(test_trace, [BasicChain])

        expected_basic = {
            'oct_fwd': 514,
            'dp': 80,
            'pkt_rev': 4,
            'pkt_fwd': 6,
            'proto': 6,
            'oct_rev': 799,
            'sip': '2001:630:241:20f:c2ea:e939:f310:9c32',
            'sp': 39956,
            'dip': '2a00:1450:4009:810::200e',
        }

        flows = self.run_observer()
        print(flows)
        assert len(flows) == 1

        for key in expected_basic:
            assert flows[0][key] == expected_basic[key]

    def test_chain_basic_ipv4_udp(self):
        test_trace = "basic_ipv4_udp.pcap"
        self.create_observer(test_trace, [BasicChain])

        expected_basic = {
            'oct_fwd': 56,
            'dp': 53,
            'pkt_rev': 1,
            'pkt_fwd': 1,
            'proto': 17,
            'oct_rev': 267,
            'sip': '172.22.152.138',
            'sp': 57606,
            'dip': '8.8.8.8'
        }

        flows = self.run_observer()
        print(flows)
        assert len(flows) == 1

        for key in expected_basic:
            assert flows[0][key] == expected_basic[key]

    def test_chain_basic_ipv6_udp(self):
        test_trace = "basic_ipv6_udp.pcap"
        self.create_observer(test_trace, [BasicChain])

        expected_basic = {
            'oct_fwd': 76,
            'dp': 53,
            'pkt_rev': 1,
            'pkt_fwd': 1,
            'proto': 17,
            'oct_rev': 287,
            'sip': '2001:470:1d58:1337:4100:e1a1:8dcf:488',
            'sp': 40672,
            'dip': '2001:4860:4860::8888',
        }

        flows = self.run_observer()
        print(flows)
        assert len(flows) == 1

        for key in expected_basic:
            assert flows[0][key] == expected_basic[key]

    def test_chain_basic_ipv4_non_udp_tcp(self):
        test_trace = "basic_ipv4_non_udp_tcp.pcap"
        self.create_observer(test_trace, [BasicChain])

        expected_basic = {
            'oct_fwd': 506,
            'dp': None,
            'pkt_rev': 0,
            'pkt_fwd': 11,
            'proto': 112,
            'oct_rev': 0,
            'sip': '192.168.0.10',
            'sp': None,
            'dip': '224.0.0.18'
        }

        flows = self.run_observer()
        print(flows)
        assert len(flows) == 1

        for key in expected_basic:
            assert flows[0][key] == expected_basic[key]
